# Import all classes for better IDE support


from _pymimir.advanced.graphs import (
    # Vertices
    EmptyVertex,
    ColoredVertex,

    # Edges
    EmptyEdge,
    ColoredEdge,

    # EmptyStaticGraph
    EmptyStaticGraph,
    ImmutableEmptyStaticGraph,
    ForwardEmptyStaticGraph,
    BidirectionalEmptyStaticGraph,

    # VertexColoredStaticGraph
    VertexColoredStaticGraph,
    ImmutableVertexColoredStaticGraph,
    ForwardVertexColoredStaticGraph,
    BidirectionalVertexColoredStaticGraph,

    # EdgeColoredStaticGraph
    EdgeColoredStaticGraph,
    ImmutableEdgeColoredStaticGraph,
    ForwardEdgeColoredStaticGraph,
    BidirectionalEdgeColoredStaticGraph,
)
